from functools import wraps
from typing import Callable

from .._autoflow.flow import Flow
from .._utility.registry import Registry
from ..models.task import PromisedTask

reg = Registry()

def flow():  # noqa: ANN201
    def wrapper(function: Callable[[], None]):  # noqa: ANN202
        @wraps(function)
        def fn(*args, **kwargs):  # noqa: ANN202, ANN002, ANN003
            with Flow() as _tree:
                __results = function(*args, **kwargs)

            if Flow.isinflow():  # if it is subflow:
                return __results
            else:
                return PromisedTask(results=__results, tree=_tree)
        return fn
    return wrapper



def parse_kv_args(raw_string: str) -> dict[str, str]:
    i = 0
    n = len(raw_string)
    in_quotes = False
    escaped = False
    buffer = ""
    key = ""
    kvargs = {}
    while i < n:
        if raw_string[i] == '"' and not escaped:
            in_quotes = not in_quotes
        elif raw_string[i] == '\\' and not escaped:
            escaped = True
        elif raw_string[i] in ' \t' and not in_quotes:
            i += 1
            continue
        elif raw_string[i] == '=' and not in_quotes:
            if buffer:
                key = buffer
                buffer = ""
            else:
                raise ValueError("Argument key cannot be empty")
        elif raw_string[i] in ', \t' and not in_quotes:
            if buffer:
                try:
                    kvargs[key.strip()] = buffer
                except ValueError:
                    raise ValueError(f"Invalid key {key.strip()}")
                key = ""
                buffer = ""
        else:
            buffer += raw_string[i]
        i += 1

    if buffer:
        kvargs[key] = buffer

    if in_quotes:
        raise ValueError("Unmatched quotes")

    return kvargs



from typing import Any, Optional

from pydantic import BaseModel

from ...constants import DEFAULT_CORE_HOST
from ...models.manifest import Dependency


class ImageOptions(BaseModel):
    image_ref: str = None
    image_auth_user: Optional[str] = None
    image_auth_pass: Optional[str] = None
    core_host: Optional[str] = DEFAULT_CORE_HOST
    core_auth_user: Optional[str] = None
    core_auth_token: Optional[str] = None
    checksum: str = None
    operations: dict[str, str] = None


class ImageDependency(Dependency):
    options: ImageOptions

    def simple(self) -> dict[str, Any]:
        simpler = {**self.options.model_dump()}
        simpler.pop("checksum", None)
        simpler.pop("operations", None)
        return {k: v for k, v in simpler.items() if v is not None}



"""
Provides functionality to install packages from plain Docker images
using Malevich Core capabilities
"""

import hashlib
import json

import malevich_coretools as core

from .._utility.host import fix_host
from ..constants import DEFAULT_CORE_HOST
from ..install.installer import Installer
from ..install.mimic import mimic_package
from ..manifest import ManifestManager
from ..models.installers.image import ImageDependency, ImageOptions

_pydantic_types = {
    "string": str.__name__,
    "integer": int.__name__,
    "number": float.__name__,
    "boolean": bool.__name__,
}


class Templates:
    """Templates for creating a package from a Docker image"""

    disclaimer = """\"\"\"
THIS FILE IS AUTOGENERATED BY malevich PACKAGE INSTALLER.

THIS FILE CONTAINS VITAL INFORMATION ABOUT THE PACKAGE AND ITS CONTENTS.
DO NOT MODIFY THIS FILE MANUALLY.
\"\"\"
"""

    imports = """
from malevich._autoflow.function import autotrace
from malevich._utility.registry import Registry
from malevich.models.nodes import OperationNode

from uuid import uuid4
from pydantic import BaseModel
"""
    registry = """
Registry().register("{operation_id}", {{
    "image_ref": {image_ref},
    "image_auth_user": {image_auth_user},
    "image_auth_pass": {image_auth_pass},
    "processor_id": "{processor_id}",
}})
"""

    schema = """
class {schema_name}(BaseModel):
{schema_def}
"""

    processor = """
@autotrace
def {name}({args}config: dict = {{}}):
    \"\"\"{docs}\"\"\"
    __instance = uuid4().hex
    return OperationNode(operation_id="{operation_id}", config=config)
"""
    init = """
# {init_name} = ...

"""

    input = """
# {input_name} = ...

"""

    output = """
# {output_name} = ...

"""


class ImageInstaller(Installer):
    """Installs package using Docker image"""
    name = "image"

    @staticmethod
    def _scan_core(
        core_auth: tuple[str, str],
        core_host: str,
        image_ref: str,
        image_auth: tuple[str, str],
    ) -> core.abstract.AppFunctionsInfo:
        """Scans Core for image info

        Args:
            core_auth: Core credentials
            core_host: Core host
            image_ref: Docker image reference
            image_auth: Docker image credentials
        """
        try:
            info = core.get_image_info(
                image_ref,
                image_auth=image_auth,
                conn_url=fix_host(core_host),
                auth=core_auth,
                parse=True
            )
        except Exception as err:
            raise Exception(
                f"Can't get info for image {image_ref}. Check that "
                "image exists and credentials are correct.\n"
                f"Error: {err}"
            )
        else:
            return info

    @staticmethod
    def create_operations(
        operations: core.abstract.AppFunctionsInfo,
        package_name: str,
    ) -> str:
        """Generates Python code for package creation

        Args:
            operations: Operations info
            package_name: Name of the package to be created
        """
        indexed_operations = {}
        salt = hashlib.sha256(operations.model_dump_json().encode()).hexdigest()
        contents = Templates.disclaimer
        contents += Templates.imports

        for schema_name, schema_fields in operations.schemes.items():
            fields = json.loads(schema_fields)["properties"]
            field_types = {
                k: _pydantic_types.get(v["type"], "str")
                for k, v in fields.items()
                if "type" in v.keys()
            }

            schema_definition = "\n".join(
                [f"\t{k}: {v} = None" for k, v in field_types.items() if k and v]
            )
            if not schema_definition:
                schema_definition = "\tpass"

            contents += Templates.schema.format(
                schema_name=schema_name,
                schema_def=schema_definition,
            )

        for id_, processor in operations.processors.items():
            args_ = []

            for arg_ in processor.arguments:
                if "return" in arg_[0] or (arg_[1] and "Context" in arg_[1]):
                    continue
                schema = None
                for name in operations.schemes.keys():
                    if arg_[1] and name in arg_[1]:
                        schema = name
                        break
                args_.append(f"{arg_[0]}{': ' + schema if schema else ''}")

            args_str_ = ", ".join(args_)
            args_str_ += ", " if args_str_ else ""

            checksum = hashlib.sha256(
                (salt + processor.model_dump_json()).encode()
            ).hexdigest()

            indexed_operations[checksum] = id_

            contents += Templates.registry.format(
                operation_id=checksum,
                image_ref=("dependencies", package_name, "options", "image_ref"),
                image_auth_user=(
                    "dependencies",
                    package_name,
                    "options",
                    "image_auth_user",
                ),
                image_auth_pass=(
                    "dependencies",
                    package_name,
                    "options",
                    "image_auth_pass",
                ),
                processor_id=id_,
            )
            contents += Templates.processor.format(
                name=id_,
                args=args_str_,
                operation_id=checksum,
                docs=processor.doc,
            )

        for id_, input_ in operations.inputs.items():
            contents += Templates.input.format(input_name=id_)

        for id_, init_ in operations.inits.items():
            contents += Templates.init.format(init_name=id_)

        for id_, output_ in operations.outputs.items():
            contents += Templates.output.format(output_name=id_)

        return contents, indexed_operations

    def install(
        self,
        package_name: str,
        image_ref: str,
        image_auth: tuple[str, str],
        core_host: str = DEFAULT_CORE_HOST,
        core_auth: tuple[str, str] = None,
    ) -> ImageDependency:
        app_info = ImageInstaller._scan_core(
            core_auth=core_auth,
            core_host=core_host,
            image_ref=image_ref,
            image_auth=image_auth,
        )
        checksum = hashlib.sha256(app_info.model_dump_json().encode()).hexdigest()
        metascript, operations = ImageInstaller.create_operations(
            app_info, package_name
        )

        mimic_package(
            package_name,
            metascript,
        )

        m = ManifestManager()
        iauth_user, iauth_pass, cauth_user, cauth_token = m.put_secrets(
            image_auth_user=image_auth[0],
            image_auth_password=image_auth[1],
            core_auth_user=core_auth[0] if core_auth else None,
            core_auth_token=core_auth[1] if core_auth else None,
            salt=checksum,
        )

        return ImageDependency(
            package_id=package_name,
            version="",
            installer="image",
            options=ImageOptions(
                checksum=checksum,
                core_host=core_host,
                core_auth_token=cauth_token,
                core_auth_user=cauth_user,
                image_auth_user=iauth_user,
                image_auth_pass=iauth_pass,
                image_ref=image_ref,
                operations=operations,
            ),
        )

    def restore(self, dependency: ImageDependency) -> None:
        core_user = dependency.options.core_auth_user
        core_token = dependency.options.core_auth_token

        return self.install(
            package_name=dependency.package_id,
            image_ref=dependency.options.image_ref,
            image_auth=(
                dependency.options.image_auth_user or "",
                dependency.options.image_auth_pass or ""
            ),
            core_host=dependency.options.core_host,
            core_auth=(
                core_user,
                core_token,
            ) if core_user and core_token else None,
        )

    def construct_dependency(self, object: dict) -> ImageDependency:
        manf = ManifestManager()
        parsed = ImageDependency(**object)
        parsed.options.image_auth_pass = manf.query_secret(
                parsed.options.image_auth_pass,
                only_value=True,
            )
        parsed.options.image_auth_user = manf.query_secret(
            parsed.options.image_auth_user,
            only_value=True,
        )
        parsed.options.core_auth_user = manf.query_secret(
            parsed.options.core_auth_user,
            only_value=True,
        )
        parsed.options.core_auth_token = manf.query_secret(
            parsed.options.core_auth_token,
            only_value=True,
        )
        return parsed



import json
import pickle
from typing import Any, Dict, List, Optional, Tuple, Union

import boto3
import jsonpickle
import numpy as np
import pandas as pd
from botocore.response import StreamingBody

WORKDIR = "/julius" # FIXME "/malevich"
APP_DIR = f"{WORKDIR}/apps"


class Context:
    class _DagKeyValue:
        """values must be bytes, string, int or float; dictionary order is not guaranteed"""  # noqa: E501
        def __init__(self, run_id: Optional[str] = None) -> None:
            pass

        def get_bytes(self, key: str) -> bytes:
            """a more optimal way to get a binary value by key (\"get\" can also be used)"""    # noqa: E501
            pass

        async def async_get_bytes(self, key: str) -> bytes:
            """a more optimal way to get a binary value by key (\"get\" can also be used)"""    # noqa: E501
            pass

        def get(self, keys: List[str]) -> Dict[str, Any]:
            pass

        async def async_get(self, keys: List[str]) -> Dict[str, Any]:
            pass

        def get_all(self) -> Dict[str, Any]:
            pass

        async def async_get_all(self) -> Dict[str, Any]:
            pass

        def update(self, keys_values: Dict[str, Any]) -> None:
            pass

        async def async_update(self, keys_values: Dict[str, Any]) -> None:
            pass

        def clear(self) -> None:
            pass

        async def async_clear(self) -> None:
            pass

    class _ObjectStorage:
        def get_keys(self, local: bool = False, all_apps: bool = False) -> List[str]:
            pass

        async def async_get_keys(self, local: bool = False, all_apps: bool = False) -> List[str]:  # noqa: E501
            pass

        def get(self, keys: List[str], force: bool = False, all_apps: bool = True) -> List[str]:    # noqa: E501
            pass

        async def async_get(self, keys: List[str], force: bool = False, all_apps: bool = True) -> List[str]:    # noqa: E501
            pass

        def get_all(self, local: bool = False, force: bool = False, all_apps: bool = True) -> List[str]:    # noqa: E501
            pass

        async def async_get_all(self, local: bool = False, force: bool = False, all_apps: bool = True) -> List[str]:    # noqa: E501
            pass

        def update(self, keys: List[str]) -> None:
            pass

        async def async_update(self, keys: List[str]) -> None:
            pass

        def delete(self, keys: List[str]) -> None:
            pass

        async def async_delete(self, keys: List[str]) -> None:
            pass

    """context for run"""
    def __init__(self) -> None:
        self.app_id: str = ""
        self.run_id: str = ""
        self.app_cfg: Dict[str, Any] = {}
        self.msg_url: str = ""
        self.email: Optional[str] = None
        self.dag_key_value = Context._DagKeyValue(self.run_id)
        self.object_storage = Context._ObjectStorage()
        self.common = None

    def share(self, path: str, all_runs: bool = False, path_prefix: str = APP_DIR, force: bool = False, synchronize: bool = True) -> None:  # noqa: E501
        """copy dir (if it doesn't already exist or \"force\"=True) or file along the path starting from the \"path_prefix\" (\"apps\" directory in app by default) to the shared directory for all apps"""  # noqa: E501
        pass

    async def async_share(self, path: str, all_runs: bool = False, path_prefix: str = APP_DIR, force: bool = False, synchronize: bool = True) -> None:  # noqa: E501
        """copy dir (if it doesn't already exist or \"force\"=True) or file along the path starting from the \"path_prefix\" (\"apps\" directory in app by default) to the shared directory for all apps"""  # noqa: E501
        pass

    def share_many(self, paths: List[str], all_runs: bool = False, path_prefix: str = APP_DIR, force: bool = False, synchronize: bool = True) -> None:  # noqa: E501
        """same as share but for multiple paths, ignore not exists path"""

    async def async_share_many(self, paths: List[str], all_runs: bool = False, path_prefix: str = APP_DIR, force: bool = False, synchronize: bool = True) -> None:  # noqa: E501
        """same as share but for multiple paths, ignore not exists path"""

    def get_share_path(self, path: str, all_runs: bool = False, not_exist_ok: bool = False) -> str:  # noqa: E501
        """return real path by path, that shared before with function \"share\""""
        pass

    def delete_share(self, path: str, all_runs: bool = False, synchronize: bool = True) -> None:    # noqa: E501
        """delete dir or file, that shared between all apps, path like path is the same as used in function \"share\""""  # noqa: E501
        pass

    async def async_delete_share(self, path: str, all_runs: bool = False, synchronize: bool = True) -> None:  # noqa: E501
        """delete dir or file, that shared between all apps, path like path is the same as used in function \"share\""""  # noqa: E501
        pass

    def synchronize(self, paths: List[str] = None, all_runs: bool = False) -> None: # TODO synchronize removing  # noqa: E501
        """synchronize mounts for pods, paths = None or [] - synchronize from root mount""" # noqa: E501
        pass

    async def async_synchronize(self, paths: List[str] = None, all_runs: bool = False) -> None:  # noqa: E501
        """synchronize mounts for pods, paths = None or [] - synchronize from root mount""" # noqa: E501
        pass

    def msg(self, data: Union[str, Dict], url: Optional[str] = None, headers: Optional[Dict[str, str]] = None, wait: bool = False, wrap: bool = True, with_result: bool = False):  # noqa: E501, ANN201
        """send http msg to system or any url

        Args:
            data (Union[str, Dict]): msg data
            url (Optional[str]): url to which the request is sent. If not specified, the default url is used (system url, but it can be overridden in the startup cfg, the parameter is msg_url). system url. Defaults to None.
            headers (Optional[Dict[str, str]]): Just headers, with which the request is sent. Defaults to None.
            wait (bool): wait result. Defaults to False.
            wrap (bool): need system wrap with operationId. Defaults to True.
            with_result (bool): return result. Defaults to False.
        """  # noqa: E501
        pass

    async def async_msg(self, data: Union[str, Dict], url: Optional[str] = None, headers: Optional[Dict[str, str]] = None, wait: bool = False, wrap: bool = True, with_result: bool = False):  # noqa: ANN201, E501
        """send http msg to system or any url

        Args:
            data (Union[str, Dict]): msg data
            url (Optional[str]): url to which the request is sent. If not specified, the default url is used (system url, but it can be overridden in the startup cfg, the parameter is msg_url). system url. Defaults to None.
            headers (Optional[Dict[str, str]]): Just headers, with which the request is sent. Defaults to None.
            wait (bool): wait result. Defaults to False.
            wrap (bool): need system wrap with operationId. Defaults to True.
            with_result (bool): return result. Defaults to False.
        """  # noqa: E501
        pass

    def email_send(self, message: str, subject: Optional[str] = None, type: str = "gmail") -> None:  # noqa: E501
        """only gmail work now, if subject is None used default"""
        pass

    def metadata(self, df_name: str) -> Optional[Union[Dict[str, Any], List[Dict[str, Any]]]]:  # noqa: E501
        """get metadata by df_name"""
        pass

    @property
    def scale_info(self) -> Tuple[int, int]:
        pass

    def get_scale_part(self, df: pd.DataFrame) -> pd.DataFrame:
        pass

    @property
    def operation_id(self) -> str:
        pass


def to_binary(smth: Any) -> bytes:  # noqa: ANN401
    return pickle.dumps(smth)


def from_binary(smth: bytes) -> Any:  # noqa: ANN401
    return pickle.loads(smth)


def load(url: str, path: str, path_prefix: str = APP_DIR) -> None:
    """path - relative: starting from the \"path_prefix\" (\"apps\" directory in app by default)"""  # noqa: E501
    pass


class S3Helper:
    def __init__(self, client: Any, s3_bucket: str) -> None:  # noqa: ANN401
        self.client = client
        self.__bucket = s3_bucket

    @staticmethod
    def create_by_cfg(cfg: Dict[str, Any], **kwargs) -> 'S3Helper':  # noqa: ANN003
        s3_client = boto3.client(
            's3',
            region_name=cfg.get('aws_region'),
            aws_access_key_id=cfg['aws_access_key_id'],
            aws_secret_access_key=cfg['aws_secret_access_key'],
            endpoint_url=cfg.get('endpoint_url'),
            **kwargs
        )
        return S3Helper(s3_client, cfg['s3_bucket'])

    def get_object(self, key: str, bucket: Optional[str]=None) -> Optional[StreamingBody]:  # noqa: E501
        pass

    def get_df(self, key: str, bucket: Optional[str]=None) -> pd.DataFrame:
        pass

    def save_object(self, body: Any, key: str, bucket: Optional[str]=None) -> None:  # noqa: ANN401
        pass

    def save_df(self, df: pd.DataFrame, key: str, bucket: Optional[str]=None) -> None:
        pass

    def delete_object(self, key: str, bucket: Optional[str]=None) -> None:
        pass


class SmtpSender:
    def __init__(self, login: str, password: str, smtp_server: str = "smtp.gmail.com", smtp_port: int = 465) -> None:  # noqa: E501
        self.server = smtp_server
        self.port = smtp_port
        self.login = login
        self.password = password

    def send(self, receivers: list[str], subject: str, message: str) -> None:
        pass


def to_df(x: Any, force: bool=False) -> pd.DataFrame:  # noqa: ANN401
    """creates a dataframe in a certain way, \"force\" should be used for complex objects ('ndarray', 'Tensor' and python primitives work without it. It crashes on basic types ('int', 'str', etc)), scheme of received dataframe - \"default_scheme\""""  # noqa: E501
    if force:
        return pd.DataFrame({"data": [jsonpickle.encode(x)]})
    elif isinstance(x, (np.ndarray, list, tuple, range, bytearray)):
        return pd.DataFrame({"data": x})
    elif type(x).__name__ == "Tensor":
        return pd.DataFrame(x)
    elif isinstance(x, (set, frozenset)):
        return pd.DataFrame({"data": list(x)})
    elif isinstance(x, dict):
        return pd.DataFrame({"data": [json.dumps(x)]})
    else:   # int, float, complex, str, bytes, bool
        return pd.DataFrame({"data": [x]})


# TODO create same with pyspark
def from_df(x: pd.DataFrame, type_name: Optional[str] = None, force: bool=False) -> Any:  # noqa: ANN401
    """decodes the \"to_df\" data from the dataframe, \"force\" is used if it was used in the encoding function - \"to_df\". You should specify the type (by type_name: for example 'ndarray', 'list', 'Tensor', 'int') that was put in this \"to_df\" dataframe.
    possible type_names: 'ndarray', 'list', 'tuple', 'Tensor', 'set', 'frozenset', 'dict', 'bytearray'. Otherwise considered a primitive base type
    if force==True ignore type_name anyway"""  # noqa: E501
    if force:
        return jsonpickle.decode(x.data[0])
    elif type_name == 'ndarray':
        return x.data.values
    elif type_name == 'list':
        return x.data.values.tolist()
    elif type_name == 'tuple':
        return tuple(x.data.values.tolist())
    elif type_name == 'range':
        return x.data.values.tolist()
    elif type_name == 'Tensor':
        import torch  # not in requirements
        return torch.from_numpy(x.values).float().to(torch.device('cpu'))   # can't work with gpu from inside yet  # noqa: E501
    elif type_name == 'set':
        return set(x.data.values.tolist())
    elif type_name == 'frozenset':
        return frozenset(x.data.values.tolist())
    elif type_name == 'dict':
        return json.loads(x.data[0])
    elif type_name == 'bytearray':
        return bytearray(x.data)
    else:   # int, float, complex, str, bytes, bool
        return x.data[0]



from abc import abstractmethod
from copy import deepcopy
from typing import Generic, Iterable, TypeVar

import pandas as pd

from .._autoflow.tracer import traced
from .._autoflow.tree import ExecutionTree
from ..models.nodes import BaseNode, TreeNode
from ..models.task import InterpretedTask

State = TypeVar("State")
Return = TypeVar("Return")


class Interpreter(Generic[State, Return]):
    """Base class for Interpreters

    An interpreter has an inner state, which is updated during the interpretation
    process. The state is updated by calling `update_state` method.

    The state is preserved in the history, which can be accessed by `history` property.
    Each time the state is updated, it is fully copied and appended to the history.

    The interpretation process can be divided into 5 steps:
    1. `before_interpret` - called before the interpretation process starts
    2. `create_node` - called when a new node is created
    3. `create_dependency` - called when a new dependency is created
    4. `after_interpret` - called after the interpretation process ends
    5. `get_result` - called to get the result of the interpretation process

    The base interpreter does not implement any of these steps. It is up to the
    child classes to implement them.

    But the base interpreter does implement the `interpret` method, which calls
    all the steps in the right order. It traverses the execution tree and calls
    the steps for each node and dependency. Traversing is preceeded by calling
    `before_interpret` and followed by calling `after_interpret`. The result of
    the interpretation process is returned by `get_result`.

    To implement a new interpreter, you need to inherit from this class and
    follow these guidelines:

    1.  Put any pipeline initialization logic into `before_interpret` method. For
        example intializiation of state, registry objects, establishing connections
        and so on.

    2.  The interpreter will traverse the execution tree and call `create_node` for
        each node and `create_dependency` for each dependency. The `create_node` method
        will be called twice for each node: first for the callee and then for the caller
        The `create_dependency` method will be called once for each dependency.

    3.  Put any pipeline finalization logic into `after_interpret` method. For example
        closing connections, saving state and so on.

    4. The result of the interpretation process is returned by `get_result` method.
        For example, you can return the state, or some part of it, or some other object.


    IMPORTANT: Ensure that the state is immutable. The state is copied and appended
    to the history each time it is updated. If the state is mutable, it will be
    changed in the history as well. This will lead to unexpected behavior.

    IMPORTANT: Ensure that ALL stages do return the state object if it is changed. The
    results of each stage are used as input for the next stage. If the state is not
    returned, the next stage will receive the old state.
    """

    def __init__(self, initial_state: State = None) -> None:
        """Initializes the interpreter

        Args:
            initial_state (State, optional): Initial state. Defaults to None.
        """
        self._state = initial_state
        self.__history = []
        self._tree = None

    def _unwrap_tree(
        self,
        tree: ExecutionTree[traced[BaseNode]]
    ) -> ExecutionTree[traced[BaseNode]]:
        """Merges all subtrees into one tree

        Args
            tree (ExecutionTree): Tree of any depth

        Returns:
            ExecutionTree: Unwrapped tree
        """
        edges = []
        for edge in tree.traverse():
            if isinstance(edge, TreeNode):
                edges.extend(self._unwrap_tree(edge).tree)
            else:
                edges.append(edge)
        return ExecutionTree(edges)

    @property
    def state(self) -> State:
        """Returns the current state

        Returns:
            State: Copy of the current state
        """
        return deepcopy(self._state)

    def update_state(self, state: State = None) -> None:
        """Updates the state

        Args:
            state (State, optional): State to update.
            If not provided, the current state is used. Defaults to None.
        """
        self.__history.append(self.state)
        if state is not None:
            self._state = deepcopy(state)
        else:
            self._state = self.state

    def interpret(self, tree: ExecutionTree[traced[BaseNode]]) -> InterpretedTask:
        """Interprets the execution tree

        The interpretation process is divided into 5 steps:
        1. `before_interpret` - called before the interpretation process starts
        2. `create_node` - called when a new node is created
        3. `create_dependency` - called when a new dependency is created
        4. `after_interpret` - called after the interpretation process ends
        5. `get_task` - called to get the executable of the interpretation process

        Before starting of the interpretation the interpreted tree is saved in the
        `_tree` property.

        Args:
            tree (ExecutionTree): Execution tree to interpret

        Returns:
            Task: Executable result of the interpretation
        """
        self._tree = self._unwrap_tree(tree)

        self.update_state(self.before_interpret(self.state))

        for from_, to, link in tree.traverse():
            self.update_state(self.create_node(self.state, from_))

            self.update_state(self.create_node(self.state, to))

            self.update_state(self.create_dependency(self.state, from_, to, link))

        self.update_state(self.after_interpret(self.state))

        return self.get_task(self.state)

    @abstractmethod
    def before_interpret(self, state: State) -> State:
        """Called before the interpretation process starts

        IMPORTANT: Ensure that the state is immutable. The state is copied and appended
        to the history each time it is updated. If the state is mutable, it will be
        changed in the history as well. This will lead to unexpected behavior.

        IMPORTANT: Ensure that the method returns the state object if it is changed. The
        results of this method are used as input for the next method.

        Args:
            state (State): Current state

        Returns:
            State: Updated state
        """
        pass

    @abstractmethod
    def create_node(
        self,
        state: State,
        tracer: traced[BaseNode]
    ) -> State:
        """Called when a new node is created

        IMPORTANT: Ensure that the state is immutable. The state is copied and appended
        to the history each time it is updated. If the state is mutable, it will be
        changed in the history as well. This will lead to unexpected behavior.

        IMPORTANT: Ensure that the method returns the state object if it is changed. The
        results of this method are used as input for the next method.

        Args:
            state (State): Current state
            tracer (tracer): A node object wrapped in a tracer object

        Returns:
            State: Updated state
        """
        pass

    @abstractmethod
    def create_dependency(
        self,
        state: State,
        callee: traced[BaseNode],
        caller: traced[BaseNode],
        link: str,
    ) -> State:
        """Called when a new dependency is created

        IMPORTANT: Ensure that the state is immutable. The state is copied and appended
        to the history each time it is updated. If the state is mutable, it will be
        changed in the history as well. This will lead to unexpected behavior.

        IMPORTANT: Ensure that the method returns the state object if it is changed. The
        results of this method are used as input for the next method.

        Args:
            state (State): Current state
            callee (tracer): A source node object wrapped in a tracer object
            caller (tracer): A destination node object wrapped in a tracer object
            link (Any): Dependency link

        Returns:
            State: Updated state
        """
        pass

    @abstractmethod
    def after_interpret(self, state: State) -> State:
        """Called after the interpretation process ends

        IMPORTANT: Ensure that the state is immutable. The state is copied and appended
        to the history each time it is updated. If the state is mutable, it will be
        changed in the history as well. This will lead to unexpected behavior.

        IMPORTANT: Ensure that the method returns the state object if it is changed. The
        results of this method are used as input for the next method.

        Args:
            state (State): Current state

        Returns:
            State: Updated state
        """
        pass

    @abstractmethod
    def get_task(self, state: State) -> InterpretedTask[State]:
        """Called after the interpretation process ends

        Args:
            state (State): Current state

        Returns:
            Task: Task to run
        """

    @abstractmethod
    def get_results(
        self,
        returned: Iterable[traced[BaseNode]] | traced[BaseNode] | None
    ) -> Iterable[pd.DataFrame] | pd.DataFrame:
        """Returns the result of the interpretation process

        Args:
            state (State): Current state
        """
        pass





from typing import Iterable

from pydantic import ConfigDict

from ..._autoflow.tracer import traced
from ..._autoflow.tree import ExecutionTree
from .base import BaseNode


class TreeNode(BaseNode):
    tree: ExecutionTree[BaseNode]
    # FIXME: Circular import with FlowOutput
    results: Iterable[traced[BaseNode]] | traced[BaseNode] | None = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True
    )



from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

from ..._autoflow.tree import ExecutionTree

State = TypeVar("State")

class AbstractSummary(ABC, Generic[State]):
    def __init__(self, tree: ExecutionTree, interpreter_state: State) -> None:
        super().__init__()

        self._tree = tree
        self._interpreter_state = interpreter_state

    @abstractmethod
    def display(self) -> None:
        pass

    @abstractmethod
    def json(self) -> dict[str, Any]:
        pass



import concurrent.futures
import logging
from typing import Annotated

import pydantic_yaml as pdyml
import rich
import typer
from malevich_space.cli.cli import app as space_app
from malevich_space.schema import SpaceSetup
from pydantic import ValidationError
from rich.progress import Progress, SpinnerColumn, TextColumn

from ._utility.args import parse_kv_args
from .commands.manifest import app as manifest_app
from .commands.use import install_from_image
from .commands.use import use as use_app
from .constants import APP_HELP
from .install.image import ImageInstaller
from .install.installer import Installer
from .manifest import ManifestManager
from .models.manifest import Dependency

logging.getLogger("gql.transport.aiohttp").setLevel(logging.ERROR)
app = typer.Typer(help=APP_HELP, rich_markup_mode="rich")

__With_Args_Help = (
    "Arguments to pass to the installer. "
    "Must be in the format [code]key1=value1,key2=value2[/code] "
    '(" can be used to escape spaces)'
)


@app.command("install", help="Install multiple packages using the image installer")
def auto_use(
    package_names: Annotated[list[str], typer.Argument(...)],
    using: Annotated[str, typer.Option(help="Installer to use")] = "image",
    with_args: Annotated[str, typer.Option(help=__With_Args_Help)] = "",
) -> None:
    rich.print(
        "\n\nAttempting automatic installation of packages [b blue]"
        f"{'[white], [/white]'.join(package_names)}[/b blue]\n"
    )

    args = parse_kv_args(with_args)

    if args:
        rich.print(
            "\n".join(
                [
                    f"\t[green_yellow]{key}[/green_yellow]: [white]{value}[/white]"
                    for key, value in args.items()
                ]
            )
        )
        rich.print("\n\n")

    with Progress(SpinnerColumn(), TextColumn("{task.description}")) as progress:
        if using == "image":
            for package_name in package_names:
                task = progress.add_task(
                    f"Attempting to install [b blue]{package_name}[/b blue] with"
                    " [i yellow]image[/i yellow] installer",
                    total=1,
                )
                try:
                    if "package_name" not in args:
                        args["package_name"] = package_name
                    install_from_image(**args)
                    args.pop("package_name")
                except Exception as err:
                    progress.update(
                        task,
                        description="[red]✘[/red] Failed with "
                        "[yellow]image[/yellow] installer "
                        f"[blue]({package_name})[/blue]",
                        completed=1,
                    )
                    progress.stop()
                    rich.print("\n\n[red]Installation failled[/red]")
                    rich.print(err)
                    exit(-1)
                else:
                    progress.update(
                        task,
                        description="[green]✔[/green] Success with "
                        f"[yellow]image[/yellow][blue] ({package_name})[/blue]",
                        completed=1,
                    )
            progress.stop()
            rich.print(
                f"\n\nInstallation of [blue]{'[white], [/white]'.join(package_names)}"
                "[/blue] with [yellow]image[/yellow] installer was "
                "[green]successful[/green]"
            )
        else:
            rich.print(
                f"[red]Installer [yellow]{using}[/yellow] is not supported[/red]"
            )
            exit(-1)


def _restore(installer: Installer, depedency: Dependency, progress: Progress) -> None:
    try:
        package_id = depedency["package_id"]
        parsed = installer.construct_dependency(depedency)
        task = progress.add_task(f"Package [green]{package_id}[/green] with "
                                f"[yellow]{installer.name}[/yellow]", total=1)
        installer.restore(parsed)
        progress.update(
            task,
            completed=True,
            description=f"[green]✔[/green] Package [b blue]{package_id}[/b blue]",
        )
    except Exception as e:
        progress.update(
            task,
            completed=True,
            description=f"[red]✘[/red] Package [b blue]{package_id}[/b blue]",
        )
        return e, package_id


@app.command(name="restore")
def restore() -> None:
    manf = ManifestManager()
    image_installer = ImageInstaller()
    with Progress(
        SpinnerColumn(), TextColumn("{task.description}")
    ) as progress, concurrent.futures.ThreadPoolExecutor() as executor:
        futures = []
        for record in manf.query("dependencies"):
            key = [*record.keys()][0]
            dependency = record[key]
            installed_by = dependency["installer"]
            if installed_by == "image":
                futures.append(
                    executor.submit(_restore, image_installer, dependency, progress)
                )
        for future in concurrent.futures.as_completed(futures):
            result = future.result()
            if result:
                progress.stop()
                rich.print(f"[red]Failed to restore package {result[1]}[/red]")
                rich.print(result[0])
                exit(-1)


@space_app.command()
def init(path_to_setup: Annotated[str, typer.Argument(...)]) -> None:
    try:
        setup = pdyml.parse_yaml_file_as(SpaceSetup, path_to_setup)
    except ValidationError as err:
        rich.print(
            f"[b red]Setup file [white]{path_to_setup}[/white] is not a correct configuration[/b red]\n"  # noqa: E501
        )  # noqa: E501
        for _err in err.errors():
            rich.print(f"\t- {_err['msg']} at [pink]{_err['loc'][0]}[pink]")
        rich.print("\n[red]Could not initialize Space installer[/red]")
    else:
        manf = ManifestManager()

        setup.password = manf.put_secret("space_password", setup.password)
        ManifestManager().put("space", value=setup)

        rich.print(
            "\nMalevich Space configuration [green]successfully[/green] added to the manifest\n"  # noqa: E501
        )  # noqa: E501


app.add_typer(use_app, name="use")
app.add_typer(manifest_app, name="manifest")
app.add_typer(space_app, name="space")


def main() -> None:
    app()


if __name__ == "__main__":
    main()



from typing import Any, ForwardRef, Generic, List, Optional, TypeVar, Union, _tp_cache

import pandas as pd
from typing_extensions import TypeVarTuple, Unpack

Schemes = TypeVarTuple('Schemes')
Scheme = TypeVar('Scheme')
SchemeM = TypeVar('SchemeM')


def _is_M(type: Any) -> bool:  # noqa: N802, ANN401
    return hasattr(type, "__origin__") and type.__origin__.__module__ == __name__ and type.__origin__.__name__ == "M"  # noqa: E501


class M(Generic[SchemeM]):
    pass


class DF(Generic[Scheme], pd.DataFrame):
    def __init__(self, df: pd.DataFrame) -> None:
        super().__init__(df)

    def cast(self, scheme: str) -> None:
        raise NotImplementedError("scheme cast not yet implemented")    # TODO

    def scheme(self) -> None:
        raise NotImplementedError("scheme not yet implemented")         # TODO

    @_tp_cache
    def scheme_name(self) -> Optional[str]:
        scheme = self._scheme_cls()
        if scheme is None:
            return None
        if hasattr(scheme, "__name__"):
            return scheme.__name__
        if hasattr(scheme, "_name"):
            return scheme._name
        if isinstance(scheme, str):
            return scheme
        if isinstance(scheme, ForwardRef):
            return scheme.__forward_arg__
        return scheme

    @_tp_cache
    def _scheme_cls(self) -> Optional[Any]:  # noqa: ANN401
        if hasattr(self, "__orig_class__"):
            return self.__orig_class__.__args__[0]
        return None


class DFS(Generic[Unpack[Schemes]]):
    def __init__(self) -> None:
        """set dfs with init"""
        self.__dfs: List[Union[DF, DFS, None]] = []
        self.__inited = False

    def init(self, *dfs: pd.DataFrame, nested: bool = False) -> 'DFS':
        """must be called after __init__, nested should be False"""
        assert not self.__inited, "DFS already inited"
        self.__inited = True
        self.__init(list(dfs), nested)
        return self

    def __init(self, dfs: List[pd.DataFrame], nested: bool = False) -> None:
        types = self.__orig_class__.__args__ if hasattr(self, "__orig_class__") else [Any for _ in dfs] # noqa: E501
        many_df_index = None
        for i, type in enumerate(types):
            if _is_M(type):
                assert not nested, "nested M in DFS"
                if many_df_index is not None:
                    raise Exception("more than one M in DFS")
                else:
                    many_df_index = i
        if many_df_index is None:
            assert len(types) == len(dfs), f"wrong arguments size: expected {len(types)}, found {len(dfs)}" # noqa: E501
            for df, type in zip(dfs, types):
                self.__dfs.append(DF[type](df))
        else:
            assert len(types) - 1 <= len(dfs), f"wrong arguments size: expected at least {len(types) - 1}, found {len(dfs)}"    # noqa: E501
            for df, type in zip(dfs[:many_df_index], types[:many_df_index]):
                self.__dfs.append(DF[type](df))
            count = len(dfs) + 1 - len(types)
            if count != 0:
                type_many = types[many_df_index].__args__[0]
                temp = DFS[tuple([type_many] * count)]().init(*dfs[many_df_index:many_df_index + count], nested=True)  # noqa: E501
                self.__dfs.append(temp)
            else:
                self.__dfs.append(None)
            for df, type in zip(dfs[many_df_index + count:], types[many_df_index + 1:]):
                self.__dfs.append(DF[type](df))

    def __len__(self) -> int:
        return len(self.__dfs)

    def __getitem__(self, key: int) -> Union[DF, 'DFS', None]:
        return self.__dfs[key]

    def __iter__(self) -> Union[DF, 'DFS', None]:
        return iter(self.__dfs)



from typing import Callable, Generic, Iterable, TypeVar, Union

import pandas as pd

from ..types import FlowOutput
from .base import BaseTask

State = TypeVar("State")

class InterpretedTask(Generic[State], BaseTask):
    def __init__(
       self,
       prepare: Callable[['InterpretedTask'], State],
       run: Callable[['InterpretedTask'], None],
       stop: Callable[['InterpretedTask'], None],
       results: Callable[['InterpretedTask', FlowOutput], Union[Iterable[pd.DataFrame], pd.DataFrame, None]],  # noqa: E501
       state: State,
       returned: FlowOutput = None,
   ) -> None:
         self.__prepare = prepare
         self.__run = run
         self.__stop = stop
         self.__results = results
         self.__state = state
         self.__returned = returned

    def prepare(self) -> None:
        self.__prepare(self)

    def run(self) -> None:
        self.__run(self)

    def stop(self) -> None:
        self.__stop(self)

    def results(self):  # noqa: ANN201
        return self.__results(self, self.__returned)

    def commit_returned(self, returned: FlowOutput):  # noqa: ANN201
        self.__returned = returned

    def __call__(self) -> Union[Iterable[pd.DataFrame], pd.DataFrame, None]:
        self.prepare()
        self.run()
        self.stop()
        return self.results()

    @property
    def state(self) -> State:
        return self.__state



import uuid
from collections import defaultdict
from typing import Any, Iterable

import malevich_coretools as core
import pandas as pd

from .._autoflow.tracer import traced
from .._utility.registry import Registry
from ..constants import DEFAULT_CORE_HOST
from ..interpreter.abstract import Interpreter
from ..manifest import ManifestManager
from ..models.collection import Collection
from ..models.nodes import BaseNode, CollectionNode, OperationNode, TreeNode
from ..models.task import InterpretedTask


class CoreInterpreterState:
    """State of the CoreInterpreter"""

    def __init__(self) -> None:
        # Involved operations
        self.ops: dict[str, BaseNode] = {}
        # Dependencies (same keys as in self.ops)
        self.depends: dict[str, list[tuple[BaseNode, tuple[str, int]]]] = defaultdict(lambda: [])  # noqa: E501
        # Registry reference (just a shortcut)
        self.reg = Registry()
        # Manifest manager reference (just a shortcut)
        self.manf = ManifestManager()
        # Task configuration
        self.cfg = core.Cfg()
        # Collections (key: operation_id, value: (`Collection` object, core_id,))
        self.collections: dict[str, tuple[Collection, str]] = {}
        # Uploaded operations (key: operation_id, value: core_op_id)
        self.core_ops: dict[str, BaseNode] = {}
        # Interpreter parameters
        self.params: dict[str, Any] = {}
        # Results
        self.results: dict[str, str] = {}


class CoreInterpreter(Interpreter[CoreInterpreterState, tuple[str, str]]):
    """Inteprets the flow using Malevich Core API"""

    def __init__(
        self,
        core_auth: tuple[str, str],
        core_host: str = DEFAULT_CORE_HOST,
    ) -> None:
        super().__init__(CoreInterpreterState())
        self.__core_host = core_host
        self.__core_auth = core_auth
        self._state.params["core_host"] = core_host
        self._state.params["core_auth"] = core_auth

        self.update_state()

    def _result_collection_name(self, operation_id: str) -> str:
        return f"result-{operation_id}"

    def create_node(
        self, state: CoreInterpreterState, node: traced[BaseNode]
    ) -> CoreInterpreterState:
        state.ops[node.owner.uuid] = node.owner
        return state

    def create_dependency(
        self,
        state: CoreInterpreterState,
        callee: traced[BaseNode],
        caller: traced[BaseNode],
        link: Any,  # noqa: ANN401
    ) -> CoreInterpreterState:
        state.depends[caller.owner.uuid].append((callee.owner, link))
        return state

    def before_interpret(self, state: CoreInterpreterState) -> CoreInterpreterState:
        core.set_host_port(self.__core_host)
        core.update_core_credentials(self.__core_auth[0], self.__core_auth[1])
        core.delete_apps()
        core.delete_tasks()
        core.delete_collections()
        core.delete_cfgs()
        return state

    def after_interpret(self, state: CoreInterpreterState) -> CoreInterpreterState:
        for id, op in state.ops.items():
            if isinstance(op, OperationNode):
                extra = state.reg.get(op.operation_id, {})


                image_auth_user = state.manf.query(
                    *extra["image_auth_user"],
                    resolve_secrets=True
                )

                image_auth_pass = state.manf.query(
                    *extra["image_auth_pass"],
                    resolve_secrets=True
                )

                image_ref = state.manf.query(
                    *extra["image_ref"],
                    resolve_secrets=True
                )

                extra_colls = {}

                for node, (link, _) in state.depends[id]:
                    if isinstance(node, CollectionNode) and \
                        node.uuid in state.collections:

                        coll, uploaded_core_id = state.collections[node.uuid]
                        state.cfg.collections = {
                            **state.cfg.collections,
                            f"{coll.collection_id}": uploaded_core_id,
                        }
                        extra_colls[link] = coll.collection_id

                app_core_name = uuid.uuid4().hex + f"-{extra['processor_id']}"
                core.create_app(
                    app_id=app_core_name,
                    processor_id=extra["processor_id"],
                    image_auth=(image_auth_user, image_auth_pass),
                    image_ref=image_ref,
                    extra_collections_from=extra_colls,
                    app_cfg=op.config,
                )

                state.core_ops[op.uuid] = app_core_name
                state.cfg.app_settings.append(
                    core.AppSettings(
                        appId=app_core_name,
                        taskId=app_core_name,
                        saveCollectionsName=[self._result_collection_name(op.uuid)],
                    )
                )
                state.results[op.uuid] = self._result_collection_name(op.uuid)

            elif isinstance(op, CollectionNode):
                collection_ref = op.collection

                uploaded_core_id = core.create_collection_from_df(
                    collection_ref.collection_data,
                    collection_ref.collection_id,
                )
                state.collections[op.uuid] = (collection_ref, uploaded_core_id)
        return state

    def get_task(
        self, state: CoreInterpreterState
    ) -> InterpretedTask[CoreInterpreterState]:
        task_kwargs = []
        config_kwargs = []

        for id, core_id in sorted(
            state.core_ops.items(), key=lambda x: len(state.depends[x[0]])
        ):
            depends = state.depends[id]
            depends.sort(key=lambda x: x[1][1])
            depends = [
                state.core_ops[x[0].uuid]
                for x in depends
                if x[0].uuid in state.core_ops
            ]

            task_kwargs.append(
                {
                    'task_id': core_id,
                    'app_id': core_id,
                    'tasks_depends': depends
                }
            )

        config_kwargs = {
            'cfg_id': uuid.uuid4().hex,
            'cfg': state.cfg,
        }

        __cfg = uuid.uuid4().hex
        core.create_cfg(__cfg, state.cfg)
        leaves = [*self._tree.leaves()]

        def prepare(
            task: InterpretedTask[CoreInterpreterState], *args, **kwargs
        ) -> None:
            for _kwargs in task_kwargs:
                core.create_task(**_kwargs)

            core.create_cfg(**config_kwargs)

            task_id = task.state.core_ops[leaves[0].owner.uuid]

            task.state.params["task_id"] = core.task_prepare(
                task_id=task_id,
                cfg_id=config_kwargs['cfg_id'],
                *args,
                **kwargs
            ).operationId

        def run(task: InterpretedTask[CoreInterpreterState], *args, **kwargs) -> None:
            if "task_id" not in task.state.params:
                raise Exception("Attempt to run a task which is not prepared. "
                                "Please, run `.prepare()` first.")
            core.task_run(task.state.params["task_id"], *args, **kwargs)

        def stop(
            task: InterpretedTask[CoreInterpreterState],
            *args,  # noqa: ANN002
            **kwargs  # noqa: ANN003
        ) -> None:
            if "task_id" not in task.state.params:
                raise Exception("Attempt to run a task which is not prepared. "
                                "Please, run `.prepare()` first.")
            core.task_stop(task.state.params["task_id"], *args, **kwargs)

        def results(
            task: InterpretedTask[CoreInterpreterState],
            returned: Iterable[traced[BaseNode]] | traced[BaseNode] | None,
            *args,  # noqa: ANN002
            **kwargs  # noqa: ANN003
        ) -> Iterable[pd.DataFrame] | pd.DataFrame:
            if "task_id" not in task.state.params:
                raise Exception("Attempt to run a task which is not prepared. "
                                "Please, run `.prepare()` first.")
            return self.get_results(returned)

        return InterpretedTask(
            prepare=prepare,
            run=run,
            stop=stop,
            results=results,
            state=state,
        )

    def get_results(
        self,
        returned: Iterable[traced[BaseNode]] | traced[BaseNode] | None
    ) -> Iterable[pd.DataFrame] | pd.DataFrame:
        if not returned:
            return None

        if isinstance(returned, traced):
            returned = [returned]

        results = []
        for r in returned:
            node = r.owner
            if isinstance(node, CollectionNode):
                results.append(node.collection.collection_data)
            elif isinstance(node, OperationNode):
                collection_ids = core.get_collections_by_name(
                    self._result_collection_name(node.uuid)
                ).ownIds

                results.append(core.get_collection_to_df(
                    collection_ids[-1]
                ))
            elif isinstance(node, TreeNode):
                results.append(self.get_results(node.results))
        return results[0] if len(results) == 1 else results





import rich

from .._autoflow.tree import ExecutionTree


def print_tree_traverse(tree: ExecutionTree) -> None:
    rich.print('[b yellow]Tree Traverse[/b yellow] (Multi-source BFS):')
    for u, v, (link, arg) in tree.traverse():
        rich.print(f'\t[blue]{u.__repr__()}[/blue] --[yellow]({link}, {arg})[/yellow]--> [blue]{v.__repr__()}[/blue]')  # noqa: E501


def print_tree_edges(tree: ExecutionTree) -> None:
    rich.print('[b yellow]Tree Edges:[/b yellow]')
    for u, v, (link, arg) in tree.tree:
        rich.print(f'\t[blue]{u.__repr__()}[/blue] --[yellow]({link}, {arg})[/yellow]--> [blue]{v.__repr__()}[/blue]')  # noqa: E501





from .._utility.singleton import SingletonMeta


class Registry(metaclass=SingletonMeta):
    _registry = {}

    def register(self, key: str, value: object) -> None:
        self._registry[key] = value

    def get(self, key: str, default=None) -> object | None:  # noqa: ANN001
        return self._registry.get(key, default)

    def __getitem__(self, key: str) -> object:
        return self._registry[key]

    def merge_keys(self, key: str, value: dict) -> None:
        if key not in self._registry:
            self._registry[key] = {}
        if not isinstance(self._registry[key], dict):
            raise ValueError(f"Key {key} is not a dictionary")

        # Deep merge of dictionaries
        self._registry[key] = self.__merge_dicts(self._registry[key], value)

    def __merge_dicts(self, dict1: dict, dict2: dict) -> dict:
        for key in dict2:
            if key in dict1:
                if isinstance(dict1[key], dict) and isinstance(dict2[key], dict):
                    dict1[key] = self.__merge_dicts(dict1[key], dict2[key])
                elif isinstance(dict1[key], list) and isinstance(dict2[key], list):
                    dict1[key] = self.__merge_lists(dict1[key], dict2[key])
                else:
                    dict1[key] = dict2[key]
            else:
                dict1[key] = dict2[key]
        return dict1

    def __merge_lists(self, list1: list, list2: list) -> list:
        for item in list2:
            if item not in list1:
                list1.append(item)
        return list1





import json

import pandas as pd
import rich
from malevich_coretools import logs_clickhouse
from rich.table import Table

from ..._utility.host import fix_host
from ..._utility.summary.abstract import AbstractSummary
from ...interpreter.core import CoreInterpreterState


class CoreTimeSummary(AbstractSummary[CoreInterpreterState]):
    def __init__(self, interpreter_state: CoreInterpreterState) -> None:
        super().__init__(None, interpreter_state)

    def json(self) -> None:
        core_ids = self._interpreter_state.core_ops.values()

        logs = logs_clickhouse(
            auth=tuple(self._interpreter_state.params["core_auth"]),
            conn_url=fix_host(self._interpreter_state.params["core_host"])
        )

        parsed = json.loads(logs)
        df = pd.DataFrame(parsed['TaskAppRunTable'])
        df = df[df.app_id.isin(core_ids)]
        df.insert(
            len(df.columns),
            column="duration",
            # Duration of the operation in milliseconds
            value=(
                pd.to_datetime(df.timestamp_end) - pd.to_datetime(df.timestamp_start)
            ).astype("timedelta64[s]").astype(int),
        )

        df.set_index("app_id", inplace=True)

        return df.duration.to_json(orient="index")

    def display(self) -> None:
        summary = json.loads(self.json())
        table = Table(title="Time Summary (Core, ms)")
        table.add_column("App", justify="left", style="cyan")
        table.add_column("Duration, s", justify="right", style="magenta")
        for app_id, duration in summary.items():
            op_name = str(app_id).split("-")[1]
            table.add_row(op_name, str(duration))

        rich.print(table)



from typing import Iterable, Union

from .._autoflow.tracer import traced
from .nodes.base import BaseNode

Nodes = Iterable[BaseNode]
TracedNode = traced[BaseNode]
TracedNodes = Iterable[traced[BaseNode]]
FlowOutput = Union[TracedNode, TracedNodes, None]




import os
import tempfile

from malevich_space.ops.component_provider.base import BaseComponentProvider
from malevich_space.parser import YAMLParser
from malevich_space.schema import ComponentSchema

from ..git.clone import clone


class CIFunctions(BaseComponentProvider):
    @staticmethod
    def has_component_in_folder(
        folder: str
    ) -> tuple[ComponentSchema | None, str]:
        """Check if a folder contains a valid component

        Args:
            folder (str): Folder to check

        Returns:
            component (ComponentSchema): Component if found, None otherwise
            path (str): Path to the component
        """
        # Scan for YAML files in the folder
        component = None
        path = os.path.abspath(folder)
        parser = YAMLParser()
        for file in os.listdir(folder):
            if file.endswith(".yaml"):
                try:
                    __component = parser._parse_comp_file(
                        os.path.join(folder, file), folder
                    )
                    if component is not None:
                        raise ValueError(
                            "Multiple valid YAML files found in folder"
                        )
                    component = __component[0]
                    path = os.path.join(folder, file)
                except Exception:
                    continue

        return component, path

    @staticmethod
    def get_available_git_components(
        github_repository: str,
        github_user: str = None,
        github_token: str = None,
        branch: str = 'main'
    ) -> list[tuple[ComponentSchema, str]]:
        components = []

        with tempfile.TemporaryDirectory() as tmpdir:
            clone(
                github_repository,
                (github_user, github_token)
                if github_user and github_token else None,
                branch,
                folder=tmpdir,
            )

            # Iterate over folder within the repository
            # with depth of one
            for folder in os.listdir(tmpdir):
                if os.path.isdir(os.path.join(tmpdir, folder)):
                    component, path = CIFunctions.has_component_in_folder(
                        os.path.join(tmpdir, folder)
                    )
                    if component is not None:
                        components.append((component, path))
        return components


    def __init__(
        self,
        git_link: str,
        git_user: str = None,
        git_token: str = None,
        branch: str = 'main'
    ) -> None:
        self.comp_paths = CIFunctions.get_available_git_components(
            github_repository=git_link,
            github_user=git_user,
            github_token=git_token,
            branch=branch
        )

    def get_by_reverse_id(self, reverse_id: str) -> ComponentSchema | None:
        for component, _ in self.comp_paths:
            if component.reverse_id == reverse_id:
                return component
        return None

    def get_all(self) -> dict[str, ComponentSchema]:
        return {
            component.reverse_id: component
            for component, _ in self.comp_paths
        }



from uuid import uuid4

from pydantic import BaseModel, Field


class BaseNode(BaseModel):
    uuid: str = Field(default_factory=lambda: uuid4().hex, repr=False)

    def __eq__(self, other: "BaseNode") -> bool:
        if not isinstance(other, BaseNode):
            return False
        return self.uuid == other.uuid



from typing import Iterable

from .._utility.singleton import SingletonMeta
from .tree import ExecutionTree


class Flow(metaclass=SingletonMeta):
    """Orchestrate the execution of the pipeline"""

    # Internal representation of the execution tree
    __flow_stack = []

    def isinflow() -> bool:
        """Check if the current execution is inside a flow"""
        return len(Flow.__flow_stack) > 0

    def flow_ref() -> ExecutionTree:
        """Returns the reference to the current flow in the context"""
        return None if not Flow.isinflow() else Flow.__flow_stack[-1]


    def __enter__(self) -> ExecutionTree:
        Flow.__flow_stack.append(ExecutionTree())
        return Flow.flow_ref()


    def __exit__(self, *args: Iterable) -> None:
        Flow.__flow_stack.pop()




from typing import Annotated

import typer

from ..constants import DEFAULT_CORE_HOST, IMAGE_BASE, USE_HELP, USE_IMAGE_HELP
from ..install.image import ImageInstaller
from ..manifest import ManifestManager

use = typer.Typer(help=USE_HELP, rich_markup_mode="rich")


@use.command("image", help=USE_IMAGE_HELP)
def install_from_image(
    package_name: Annotated[str, typer.Argument(help="Package name")],
    image_ref: Annotated[
        str, typer.Argument(help="Image reference in format <registry>/<image>:<tag>")
    ] = None,
    image_auth_user: Annotated[
        str, typer.Argument(help="Image registry auth user")
    ] = "",
    image_auth_password: Annotated[
        str, typer.Argument(help="Image registry auth password")
    ] = "",
    core_host: Annotated[
        str, typer.Option(help="Malevich Core hostname")
    ] = DEFAULT_CORE_HOST,
    core_user: Annotated[str, typer.Option(help="Malevich Core user")] = None,
    core_token: Annotated[str, typer.Option(help="Malevich Core token")] = None,
) -> None:
    if image_ref is None:
        return install_from_image(
            package_name=package_name,
            image_ref=IMAGE_BASE.format(app=package_name),
            image_auth_user=image_auth_user,
            image_auth_password=image_auth_password,
            core_host=core_host or DEFAULT_CORE_HOST,
            core_user=core_user,
            core_token=core_token,
        )

    installer = ImageInstaller()
    manifest_entry = installer.install(
        package_name=package_name,
        image_ref=image_ref,
        image_auth=(image_auth_user, image_auth_password),
        core_host=core_host,
        core_auth=(core_user, core_token) if core_user and core_token else None,
    )
    manifest_manager = ManifestManager()
    if manifest_manager.query("dependencies", package_name):
        manifest_manager.put(
            "dependencies",
            package_name,
            value=manifest_entry.model_dump(),
        )
    else:
        manifest_manager.put(
            "dependencies",
            value={f"{package_name}": manifest_entry.model_dump()},
            append=True,
        )

    manifest_manager.save()



import uuid

import pandas as pd
from pydantic import BaseModel, ConfigDict, Field


class Collection(BaseModel):
    collection_id: str
    collection_data: pd.DataFrame = Field(..., repr=False)
    model_config = ConfigDict(
        arbitrary_types_allowed=True
    )

    @staticmethod
    def from_file(file: str, id: None = uuid.uuid4()) -> None:
        return Collection(
            collection_id=id,
            collection_data=pd.read_csv(file)
        )



import os
import tempfile

from git import Repo
from github import Github

from .base import CIOps


class GithubCIOps(CIOps):
    def __has_malevich_actions(self, repo_dir: str) -> bool:
        return os.path.exists(
            os.path.join(repo_dir, ".github", "workflows", "malevich-ci.yml")
        )

    def __create_action(self, repo_dir: str, branch: str) -> str:
        os.makedirs(os.path.join(repo_dir, ".github", "workflows"), exist_ok=True)

        action_file_path = os.path.join(
            repo_dir, ".github", "workflows", "malevich-ci.yml"
        )
        template_file_path = os.path.join(
            os.path.dirname(__file__), 'templates', 'malevich-ci.yml')
        with open(action_file_path, "w") as action, \
            open(template_file_path) as template:

            action.write(template.read().format(
                BRANCH_NAME=branch
            ))

        return action_file_path

    def _log(self, verbose: bool, msg: str) -> None:
        if verbose:
            print(msg)


    def setup_ci(
        self,
        token: str,
        repository: str,
        space_user: str,
        space_token: str,
        space_url: str,
        branch: str,
        registry_url: str,
        registry_id: str,
        image_user: str,
        image_token: str,
        verbose: bool = False,
    ) -> None:
        """Setup CI for a Github repository

        Args:
            token (str): Github token
            repository (str): Github repository in the form of `user/repo`
            space_user (str): Malevich Space username
            space_token (str): Malevich Space token (password)
            space_url (str): Malevich Space API URL
            branch (str): Branch to setup CI in
            registry_url (str):
                Docker Image Registry URL, for example `public.ecr.aws` or 'cr.yandex'
            registry_id (str): Docker Registry ID
            image_user (str): Username to access the Docker Image Registry
            image_token (str): TOKEN to access the Docker Image Registry
            verbose (bool, optional): Verbose mode. Defaults to False.
        """
        g = Github(token)

        repo = g.get_repo(repository)

        repo.create_secret("SPACE_USERNAME",space_user)
        self._log(verbose, "Updated secret SPACE_USERNAME")
        repo.create_secret("SPACE_PASSWORD", space_token)
        self._log(verbose, "Updated secret SPACE_PASSWORD")
        repo.create_secret("API_URL", space_url)
        self._log(verbose, "Updated secret API_URL")
        repo.create_secret("IMAGE_USERNAME", image_user)
        self._log(verbose, "Updated secret IMAGE_USERNAME")
        repo.create_secret("IMAGE_PASSWORD", image_token)
        self._log(verbose, "Updated secret IMAGE_PASSWORD")
        repo.create_secret("REGISTRY_URL", registry_url)
        self._log(verbose, "Updated secret REGISTRY_URL")
        repo.create_secret("REGISTRY_ID", registry_id)
        self._log(verbose, "Updated secret REGISTRY_ID")
        repo.create_secret('REGISTRY_TYPE', 'ecr' if 'ecr' in registry_url else 'other')
        self._log(verbose, "Updated secret REGISTRY_TYPE")

        with tempfile.TemporaryDirectory() as tmpdir:
            git_repo = Repo.clone_from(repo.clone_url, tmpdir)
            self._log(verbose, f"Cloned repository {repository} into {tmpdir}")
            head = git_repo.create_head(branch)
            head.checkout()
            self._log(verbose, f"Checkouted to branch {branch}")
            action_file = self.__create_action(tmpdir, branch)
            self._log(verbose, f"Created action file {action_file}")
            git_repo.index.add([action_file])
            git_repo.index.commit("add: malevich-ci.yml")
            self._log(verbose, f"Committed action file {action_file}")
            origin = git_repo.remote(name="origin")
            self._log(verbose, "Pushing to origin")
            origin.push(git_repo.active_branch)
            self._log(verbose, "Pushed to origin")






from ..._autoflow.tree import ExecutionTree
from ...interpreter.abstract import Interpreter
from ...models.task.base import BaseTask
from ...models.types import FlowOutput


class PromisedTask(BaseTask):
    def __init__(self, results: FlowOutput, tree: ExecutionTree) -> None:
        self.__results = results
        self.__tree = tree
        self.__task = None

    def interpret(self, interpreter: Interpreter) -> None:
        task = interpreter.interpret(self.__tree)
        task.commit_returned(self.__results)
        self.__task = task

    def prepare(self) -> None:
        if not self.__task:
            raise Exception(
                "Unable to prepare task, that has not been interpreted. "
                "Please, use `.interpret` first to attach task to "
                "a particular platform"
            )

        return self.__task.prepare()


    def run(self) -> None:
        if not self.__task:
            raise Exception(
                "Unable to run task, that has not been interpreted. "
                "Please, use `.interpret` first to attach task to "
                "a particular platform"
            )
        # TODO: try/except with error on this level
        # if task is not prepared
        return self.__task.run()


    def stop(self) -> None:
        if not self.__task:
            raise Exception(
                "Unable to prepare task, that has not been interpreted. "
                "Please, use `.interpret` first to attach task to "
                "a particular platform"
            )
        # TODO: try/except with error on this level
        # if task is not prepared
        return self.__task.stop()


    def results(self) -> None:
        if not self.__task:
            raise Exception(
                "Unable to get results of the task, that has not been interpreted. "
                "Please, use `.interpret` first to attach task to "
                "a particular platform"
            )

        return self.__task.results()


    def commit_returned(self, returned: FlowOutput) -> None:
        if not self.__task:
            raise Exception(
                "Unable to commit returned results to the task, "
                "that has not been interpreted. "
                "Please, use `.interpret` first to attach task to "
                "a particular platform"
            )

        return self.__task.commit_returned(returned)



from .._autoflow.tree import ExecutionTree
from ..constants import DEFAULT_CORE_HOST
from ..models.verdict import CanInterpretVerdict
from .abstract import Interpreter
from .core import CoreInterpreter

INTERPRETERS = [
    CoreInterpreter(
        core_host=DEFAULT_CORE_HOST,
        core_auth=None,
    )
]

def auto(tree: ExecutionTree) -> tuple[Interpreter, list[CanInterpretVerdict]]:
    verdicts = []
    interpreter = None
    for _interpreter in INTERPRETERS:
        verdict =_interpreter.can_interpret(tree)
        verdicts.append(verdict)
        if verdict.verdict:
            interpreter = _interpreter
    return interpreter, verdicts



from abc import ABC, abstractmethod

from malevich.models.types import FlowOutput


class BaseTask(ABC):
    @abstractmethod
    def prepare(self) -> None:
        return

    @abstractmethod
    def run(self) -> None:
        return

    @abstractmethod
    def stop(self) -> None:
        return

    @abstractmethod
    def results(self):  # noqa: ANN201
        return

    @abstractmethod
    def commit_returned(self, returned: FlowOutput) -> None:
        return


    def __call__(self):  # noqa: ANN204
        self.prepare()
        self.run()
        self.stop()
        return self.results()



from malevich_space.schema import SpaceSetup

from ...manifest import ManifestManager

manf = ManifestManager()

def resolve_setup(space: dict) -> SpaceSetup:
    space = SpaceSetup(**space)
    space.password = manf.query_secret(space.password, only_value=True)
    return space



from setuptools import setup

requirements = open('requirements.txt').read().split()
version = open('VERSION').read().strip()

setup(
    name="malevich",
    version=version,
    description=(
        "A user-friendly interface to communicate with Malevich AI (malevich.ai)",
    ),
    entry_points={
        'console_scripts': [
            'malevich = malevich.cli:main',
        ]
    },
    install_requires=requirements,
    package_data={"malevich": ["**/*.yml"]}
)



from abc import ABC, abstractmethod


class CIOps(ABC):
    @abstractmethod
    def setup_ci(
        self,
        user: str,
        token: str,
        repository: str,
        space_user: str,
        space_token: str,
        space_url: str,
        branch: str,
        registry_url: str,
        registry_id: str,
        image_user: str = None,
        image_token: str = None,
        verbose: bool = False,
    ) -> None:
        pass



import logging
# Autoflow Engine
from ._autoflow import tracer as gn
from ._autoflow import tree as tree
from ._autoflow.function import *
from ._autoflow.flow import *

# Metascript
from ._meta.collection import collection
from ._meta.flow import flow
from ._meta.config import config

# Manifest
from .manifest import *
from .models.manifest import *

# Expose contants
from .constants import *

# Summaries
from ._utility.summary import *

manifest = ManifestManager()


import malevich_coretools as core
__logger = logging.getLogger("log")
__logger.setLevel(logging.CRITICAL)
core.set_logger(__logger)



import functools
from typing import Callable, ParamSpec, TypeVar

from . import tracer as gn

C = ParamSpec("C")
R = TypeVar("R")


def autotrace(func: Callable[C, R]) -> Callable[C, R]:
    @functools.wraps(func)
    def wrapper(*args, **kwargs):  # noqa: ANN202, ANN002, ANN003
        result = func(*args, **kwargs)
        result = gn.traced(result) if not isinstance(result, gn.traced) else result

        for arg in args:
            argument_name = func.__code__.co_varnames[args.index(arg)]
            if isinstance(arg, gn.traced):
                arg._autoflow.calledby(result, (argument_name, args.index(arg)))
        for key in kwargs:
            if isinstance(kwargs[key], gn.traced):
                kwargs[key]._autoflow.calledby(result, (key, kwargs[key]))

        return result

    return wrapper
